/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <srf.h>
#include <hashtable.h>
#include <value.h>
#include <function.h>
#include <builtins.h>

#ident "$Id: builtins.c,v 1.12 2004/11/02 04:39:39 bediger Exp $"

struct value *successor(struct function *own, struct value_list *arglist);
struct value *zero(struct function *own, struct value_list *arglist);

struct value *identity_constructor(struct function *own, struct value_list *arglist);
struct value *identity_calculator(struct function *own, struct value_list *arglist);

struct value *composition_constructor(struct function *own, struct value_list *arglist);
struct value *composer(struct function *own, struct value_list *arglist);

struct value *recursion_constructor(struct function *own, struct value_list *arglist);
struct value *recursor(struct function *own, struct value_list *arglist);

struct value *minimization_constructor(struct function *own, struct value_list *arglist);
struct value *minimizor(struct function *own, struct value_list *arglist);

void print_arguments(char *leading_string, struct value_list *list);

int primitive_call_counter = 0;

struct value *
successor(struct function *own, struct value_list *arglist)
{
	struct value *r = new_value();

	assert(SUCCESSOR == own->type);

	++primitive_call_counter;

	if (arglist)
	{
		int integer_val = evaluate(arglist->head);
		r->type = INTEGER;
		r->v.value = integer_val + 1;
		if (own->debugflag)
			print_arguments("successor", arglist);
	} else {
		/* no arguments - an error */
		printf("Successor: no arguments\n");
		r = new_value();
	}

	if (own->debugflag)
		fprintf(stderr, "succesor returning %d\n", r->v.value);

	return r;
}

struct value *
zero(struct function *own, struct value_list *arglist)
{
	struct value *r = new_value();

	assert(ZERO == own->type);

	++primitive_call_counter;

	r->type = INTEGER;
	r->v.value = 0;

	if (own->debugflag)
		print_arguments("zero", arglist);

	return r;
}

struct value *
identity_constructor(struct function *own, struct value_list *arglist)
{
	struct value *r = new_value();
	int count, index;

	assert(CONSTRUCTOR == own->type);

	if (2 != arglist->cnt)
	{
		fprintf(stderr, "id: needs two numeric arguments\n");
		return r;
	}
	if (INTEGER != arglist->head->type)
	{
		if (IDENTIFIER != arglist->head->type)
		{
			fprintf(stderr, "id: needs two numeric argumenst\n");
			return r;
		}
		/* need to check that the IDENTIFIER names a number */
	}

	count = evaluate(arglist->head);
	index = evaluate(arglist->head->next);

	if (0 == count || 0 == index)
	{
		fprintf(stderr, "Can't have 0 count of arguments\n");
		return r;
	}

	r->type = FUNCTION;
	r->v.function = new_function();
	r->v.function->function = identity_calculator;

	r->v.function->type = IDENTITY;

	r->v.function->identity_data.count = count;
	r->v.function->identity_data.index = index;

	if (r->v.function->identity_data.index > r->v.function->identity_data.count)
	{
		printf("id[%d, %d]: fewer arguments than desired argument index\n",
			r->v.function->identity_data.count,
			r->v.function->identity_data.index
		);
		r->v.function->identity_data.index
			= r->v.function->identity_data.count;
	}
	--r->v.function->identity_data.index;

	return r;
}

void
setup_builtins(void)
{
	struct value *v;

	v = new_value();
	v->type = FUNCTION;
	v->v.function = new_function();
	v->v.function->function = zero;
	v->v.function->type = ZERO;
	v->assigned = 1;

	add_data(value_table, "z", (void *)v);

	v = new_value();
	v->type = FUNCTION;
	v->v.function = new_function();
	v->v.function->function = successor;
	v->v.function->type = SUCCESSOR;
	v->assigned = 1;

	add_data(value_table, "s", (void *)v);

	v = new_value();
	v->type = FUNCTION;
	v->v.function = new_function();
	v->v.function->function = identity_constructor;
	v->v.function->type = CONSTRUCTOR;
	v->assigned = 1;

	add_data(value_table, "id", (void *)v);

	v = new_value();
	v->type = FUNCTION;
	v->v.function = new_function();
	v->v.function->function = composition_constructor;
	v->v.function->type = CONSTRUCTOR;
	v->assigned = 1;

	add_data(value_table, "Cn", (void *)v);

	v = new_value();
	v->type = FUNCTION;
	v->v.function = new_function();
	v->v.function->function = recursion_constructor;
	v->v.function->type = CONSTRUCTOR;
	v->assigned = 1;

	add_data(value_table, "Pr", (void *)v);

	v = new_value();
	v->type = FUNCTION;
	v->v.function = new_function();
	v->v.function->function = minimization_constructor;
	v->v.function->type = CONSTRUCTOR;
	v->assigned = 1;

	add_data(value_table, "Mn", (void *)v);
}

struct value *
identity_calculator(struct function *own, struct value_list *arglist)
{
	struct value *r = new_value(), *p;
	int x, i;

	assert(IDENTITY == own->type);

	++primitive_call_counter;

	x = own->identity_data.index;

	if (arglist->cnt < own->identity_data.count)
	{
		printf("Fewer arguments (%d) than required (%d) by id[%d,%d]\n",
			arglist->cnt,
			own->identity_data.count,
			own->identity_data.count, own->identity_data.index + 1
		);
			
		return r;
	}

	for (p = arglist->head, i = 0; p && i < x; ++i, p = p->next)
		;

	r->type = INTEGER;
	r->v.value = evaluate(p);

	return r;
}

/* Implements the Cn[] functor - makes a struct value that has
 * type of FUNCTION, and a valid member function that composes
 * the values of its argument functions. */
struct value *
composition_constructor(struct function *own, struct value_list *arglist)
{
	struct value *r = new_value();
	struct value *p;
	int i;

	assert(CONSTRUCTOR == own->type);

	r->type = FUNCTION;

	r->v.function = new_function();
	r->v.function->function = composer;

	r->v.function->type = COMPOSITION;

	/* r->v.function->composition_data.first: "outermost" composed function */
	/* r->v.function->composition_data.rest: "inner" composed function(s) */

	r->v.function->composition_data.first
		= copy_function(value_function(arglist->head));
	r->v.function->composition_data.rest
		= malloc(sizeof(struct function *) * arglist->cnt);

	for (i = 0, p = arglist->head->next; p; ++i)
	{
		r->v.function->composition_data.rest[i]
			= copy_function(value_function(p));
		p = p->next;
	}

	r->v.function->composition_data.rest[i] = NULL;

	return r;
}

/* Execute a function built with Cn[] functor */
struct value *
composer(struct function *own, struct value_list *arglist)
{
	struct value *r = NULL;
	struct value_list *evaluated_list, *sub_list;
	struct value *p;
	int i;

	assert(COMPOSITION == own->type);

	/* 1. Convert arglist into a list of evaluated-as-number
	 *    structs value */
	evaluated_list = copy_and_evaluate(arglist);

	if (own->debugflag)
		print_arguments("composer", arglist);

	/* 2. Increment "assigned" field to prevent them getting disposed of
	 *    in called functions */
	evaluated_list->assigned = 1;
	p = evaluated_list->head;
	while (p)
	{
		++p->assigned;
		p = p->next;
	}

	/* 3. Call each of own->composition_data.rest members with converted
	 *    list of structs value, created a struct value_list * */
	sub_list = new_value_list();
	sub_list->head = sub_list->tail = NULL;
	sub_list->cnt = 0;
	for (i = 0; own->composition_data.rest[i]; ++i)
	{
		struct value *q
			= evaluate_function(own->composition_data.rest[i], evaluated_list);

		if (own->debugflag)
			printf("composer argument %d has value %d\n", i, q->v.value);

		/* create struct value_list of return values of the "rest"
		 * of the functions */
		if (sub_list->tail)
		{
			sub_list->tail->next = q;
			sub_list->tail = q;
		} else {
			sub_list->head = q;
			sub_list->tail = q;
		}
		++sub_list->cnt;
	}

	/* 4. Decrement "assigned" field */
	p = evaluated_list->head;
	while (p)
	{
		--p->assigned;
		p = p->next;
	}
	evaluated_list->assigned = 0;

	/* 5. Free the converted list */
	free_value_list(evaluated_list);

	/* 6. Call own->composition_data.first with the newly-create value_list */
	r =  evaluate_function(own->composition_data.first, sub_list);

	if (own->debugflag)
		printf("composer, composed function returns %d\n", r->v.value);

	/* 7. Return that struct value  from own->composition_data.first */
	return r;
}

struct value *
recursion_constructor(struct function *own, struct value_list *arglist)
{
	struct value *r = new_value();

	assert(CONSTRUCTOR == own->type);

	r->type = FUNCTION;

	r->v.function = new_function();
	r->v.function->function = recursor;

	r->v.function->type = PREC;

	r->v.function->recursion_data.zero
		= copy_function(value_function(arglist->head));
	r->v.function->recursion_data.increment
		= copy_function(value_function(arglist->head->next));

	return r;
}

/*
 * Example recursive call to illustrate the for-loop conditions:
 * add = Pr[f, g];
 *
 * add(x, 0) = f(x);
 * add(x, y') = g(x, y, add(x, y));
 *
 * add(2,3) = g(2, 2, add(2, 2))
 * add(2,2) = g(2, 1, add(2, 1))
 * add(2,1) = g(2, 0, add(2, 0))
 * add(2,0) = f(2)
 *
 * The for-loop in recursor() counts up from 0, so the 2nd argument
 * to the "increment" function has to take on values 0, 1, 2 when
 * "recursing" on the functions above.  The "increment" function
 * gets called with 2nd argument value of 0, even though add(x, 0)
 * gets calculated by calling g(x).
 */
struct value *
recursor(struct function *own, struct value_list *arglist)
{
	struct value_list *vl;
	struct value *zv, *r = NULL;
	struct value *tl;
	int i, max;

	assert(PREC == own->type);

	vl = copy_and_evaluate(arglist);
	++vl->assigned;

	if (own->debugflag)
		print_arguments("recursor", vl);

	zv = evaluate_function(own->recursion_data.zero, vl);

	max = evaluate(vl->tail);

	if (own->debugflag)
		fprintf(stderr, "recursor, zero value %d, max value %d\n",
			zv->v.value, max);

	tl = vl->tail;
	tl->next = zv;
	vl->tail = zv;
	++vl->cnt;

	/* an interative for-loop can perform the recursion
	 * because of the way that we state Pr[].  Basically,
	 * this for-loop does "tail recursion optimization". */
	for (i = 0; i < max; ++i)
	{
		struct value *t;

		tl->v.value = i;

		if (own->debugflag)
		{
			char buf[256];
			sprintf(buf, "recursor, iteration %d", i);
			print_arguments(buf, vl);
		}

		t = evaluate_function(own->recursion_data.increment, vl);

		free_value(vl->tail);
		tl->next = t;
		vl->tail = t;
	}

	r = vl->tail;

	vl->tail = tl;
	vl->tail->next = NULL;
	--vl->cnt;

	--vl->assigned;
	free_value_list(vl);

	if (own->debugflag)
		fprintf(stderr, "recursor returning %d\n", r->v.value);

	return r;
}

void
print_arguments(char *leading_string, struct value_list *list)
{
	struct value *tmp;
	fprintf(stderr, "%s: Argument list length %d%c ",
		leading_string, list->cnt, list->cnt? ':': ' ');
	for (tmp = list->head; tmp; tmp = tmp->next)
		fprintf(stderr, "%d%s",
			evaluate(tmp), tmp->next? ", ": "");
	fprintf(stderr, "\n");
}

struct value *
minimization_constructor(struct function *own, struct value_list *arglist)
{
	struct value *r = new_value();

	assert(CONSTRUCTOR == own->type);

	r->type = FUNCTION;

	r->v.function = new_function();
	r->v.function->function = minimizor;

	r->v.function->type = MIN;

	/* XXX - need to ensure that arglist has a member: a function */
	r->v.function->minimization_data.function
		= copy_function(value_function(arglist->head));

	return r;
}

/*
 * Minimization - calculates
 * Mn[f], where f constitutes a function taking N arguments.
 * Called like Mn[f](a0, a1, a2, ... aM), where M == N - 1
 * This function has to put in argument aN and values for
 * argument aN ranging from 0 until f(a0, a1, a2,... aN)
 * evalutes to 0.
 */
struct value *
minimizor(struct function *own, struct value_list *arglist)
{
	struct value_list *vl = NULL;
	struct value *tl = NULL, *r = new_value();
	int i = 0, looping = 1;

	assert(MIN == own->type);

	vl = copy_and_evaluate(arglist);
	++vl->assigned;

	if (own->debugflag)
		print_arguments("minimizor", vl);

	r->type = INTEGER;
	++r->assigned;

	tl = vl->tail;
	if (vl->tail)
		vl->tail->next = r;
	if (!vl->head)
		vl->head = r;
	vl->tail = r;
	++vl->cnt;

	while (looping)
	{
		struct value *t = NULL;

		vl->tail->v.value = i;

		if (own->debugflag)
		{
			char buf[256];
			sprintf(buf, "minimizor, iteration %d", i);
			print_arguments(buf, vl);
		}

		t = evaluate_function(own->minimization_data.function, vl);

		looping = (0 != t->v.value);

		free_value(t);
		t = NULL;

		++i;
	}

	vl->tail = tl;
	if (vl->tail)
		vl->tail->next = NULL;
	if (r == vl->head)
		vl->head = NULL;
	--vl->cnt;

	--vl->assigned;
	free_value_list(vl);

	--r->assigned;

	if (own->debugflag)
		fprintf(stderr, "minimizor returning %d\n", r->v.value);

	return r;
}
