%{
/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ident "$Id: lex.l,v 1.5 2004/11/02 04:39:45 bediger Exp $"
#include <stdlib.h>
#include <string.h>

#include "rfunc-gram2.h"

int lineno = 1;
%}

%%

\n		{ ++lineno; }

[ \t]	{ /* ignore whitespace */ }

"0x"[0-9A-Fa-f]+ {
	/* Numeric constants: 0xXXX hexadecimal,
 	* 0XXX octal, regular decimal and 0101b binary. */
		yylval.con = strtoul(yytext, NULL, 0x10);
		return TK_CONST;
	}
"0"[0-7]+ {
		yylval.con = strtoul(yytext, NULL, 010);
		return TK_CONST;
	}
[0-9]+	{  yylval.con = strtoul(yytext, NULL, 10); return TK_CONST; }
[01]+"b" { yylval.con = strtoul(yytext, NULL, 2);  return TK_CONST;  /* this seems crappy - trust strtoul() to ingnore the trailing 'b' */ }

"if"   { return TK_IF; }
"else" { return TK_ELSE; }


[a-zA-Z_][a-zA-Z0-9_]* {
	/* identifiers */
	yylval.ident = malloc(strlen(yytext) + 1);
	strcpy((char *)yylval.ident, yytext);
	return TK_IDENTIFIER;
}

\+		{ return TK_PLUS; }
\-		{ return TK_MINUS; }
\*		{ return TK_TIMES; }
\/		{ return TK_DIV; }
\(		{ return TK_LPAREN; }
\)		{ return TK_RPAREN; }
\;		{ return TK_SEMI; }
\,		{ return TK_COMMA; }
\<		{
		yylval.ident = malloc(strlen(yytext) + 1);
		strcpy((char *)yylval.ident, yytext);
		return TK_LOGICAL_OP;
	}
\>		{
		yylval.ident = malloc(strlen(yytext) + 1);
		strcpy((char *)yylval.ident, yytext);
		return TK_LOGICAL_OP;
	}
"<="	{
		yylval.ident = malloc(strlen(yytext) + 1);
		strcpy((char *)yylval.ident, yytext);
		return TK_LOGICAL_OP;
	}
">="	{
		yylval.ident = malloc(strlen(yytext) + 1);
		strcpy((char *)yylval.ident, yytext);
		return TK_LOGICAL_OP;
	}
"=="	{
		yylval.ident = malloc(strlen(yytext) + 1);
		strcpy((char *)yylval.ident, yytext);
		return TK_LOGICAL_OP;
	}
"!="	{
		yylval.ident = malloc(strlen(yytext) + 1);
		strcpy((char *)yylval.ident, yytext);
		return TK_LOGICAL_OP;
	}
.		{ /* Just eat it up */ }

%%

/* when the lexer hits end-of-file, it calls yywrap() - if yywrap()
 * returns 0, the lexer tries again.  If yywrap() returns 1, the lexer
 * exits, which seems to cause yyparse() (yacc or bison generated) to exit.
 */
int
yywrap()
{
	return 1;
}
